/*
    CHAOS target for netfilter

    Written by Jan Engelhardt, 2006 - 2007
    released under the terms of the GNU General Public
    License version 2.x and only versions 2.x.
*/
#include <linux/icmp.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/stat.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter/xt_tcpudp.h>
#include <linux/netfilter/xt_statistic.h>
#include <linux/netfilter_ipv4/ipt_REJECT.h>
#include <net/ip.h>
#include "xt_CHAOS.h"
#include "find_match.c"
#define PFX KBUILD_MODNAME ": "

/* Module parameters */
static int reject_percentage = .01 * 0x80000000;
module_param(reject_percentage, int, S_IRUGO | S_IWUSR);

/* Static data for other matches/targets */
static const struct ipt_reject_info reject_params = {
	.with = ICMP_HOST_UNREACH,
};

static const struct xt_tcp tcp_params = {};

/* References to other matches/targets */
static struct xt_match *xm_tcp;
static struct xt_target *xt_reject, *xt_tarpit;

/* CHAOS functions */
static unsigned int xt_chaos_target(struct sk_buff **pskb,
    const struct net_device *in, const struct net_device *out,
    unsigned int hooknum, const struct xt_target *target, const void *targinfo,
    void *userinfo)
{
	/* Equivalent to:
	 * -A chaos -m random --average 1 -j REJECT --reject-with host-unreach
	 * -A chaos -p tcp -j TARPIT
	 * -A chaos -j DROP
	 */
	const struct xt_chaos_info *info = targinfo;

	if(((int)net_random() & 0x7FFFFFFF) <= reject_percentage)
		return xt_reject->target(pskb, in, out, hooknum, target,
		       &reject_params, userinfo);

	/* TARPIT may not be called from the OUTPUT chain */
	if(info->tarpit && hooknum != NF_IP_LOCAL_OUT &&
	  (*pskb)->nh.iph->protocol == IPPROTO_TCP)
	{
		const int protoff = 4 * (*pskb)->nh.iph->ihl;
		const int offset = ntohs((*pskb)->nh.iph->frag_off) & IP_OFFSET;
		int hotdrop = 0, ret;

		ret = xm_tcp->match(*pskb, in, out, xm_tcp, &tcp_params,
		                    offset, protoff, &hotdrop);
		if(ret && !hotdrop)
			return xt_tarpit->target(pskb, in, out, hooknum,
			       target, NULL, userinfo);
	}

	return NF_DROP;
}

static int xt_chaos_check(const char *tablename, const void *entry,
    const struct xt_target *target, void *targinfo, unsigned int targinfosize,
    unsigned int hook_mask)
{
	const struct xt_chaos_info *info = targinfo;
	if(targinfosize != XT_ALIGN(sizeof(struct xt_chaos_info))) {
		printk(KERN_WARNING PFX "targinfosize %u != %Zu\n",
		       targinfosize, XT_ALIGN(sizeof(struct xt_chaos_info)));
		return 0;
	}
	if(info->tarpit & ~1) {
		printk(KERN_WARNING PFX "Invalid value for tarpit param\n");
		return 0;
	}
	return 1;
}

static struct xt_target xt_CHAOS = {
	.name       = "CHAOS",
	.target     = xt_chaos_target,
	.checkentry = xt_chaos_check,
	.table      = "filter",
	.targetsize = sizeof(struct xt_chaos_info),
	.hooks      = (1 << NF_IP_LOCAL_IN) | (1 << NF_IP_FORWARD) |
	              (1 << NF_IP_LOCAL_OUT),
	.family     = AF_INET,
	.me         = THIS_MODULE,
};

static int __init xt_chaos_init(void)
{
	int ret = -EINVAL;

	if((xm_tcp = xt_request_find_match(AF_INET, "tcp", 0)) == NULL)
		return -EINVAL;
	if((xt_reject = xt_request_find_target(AF_INET, "REJECT", 0)) == NULL)
		goto out2;
	if((xt_tarpit = xt_request_find_target(AF_INET, "TARPIT", 0)) == NULL)
		goto out3;
	if((ret = xt_register_target(&xt_CHAOS)) != 0)
		goto out4;
	return 0;

 out4:
	module_put(xt_tarpit->me);
 out3:
	module_put(xt_reject->me);
 out2:
	module_put(xm_tcp->me);
	return ret;
}

static void __exit xt_chaos_exit(void)
{
	xt_unregister_target(&xt_CHAOS);
	module_put(xm_tcp->me);
	module_put(xt_reject->me);
	module_put(xt_tarpit->me);
	return;
}

module_init(xt_chaos_init);
module_exit(xt_chaos_exit);
MODULE_AUTHOR("Jan Engelhardt <jengelh@gmx.de>");
MODULE_DESCRIPTION("netfilter CHAOS target module");
MODULE_LICENSE("GPL");
MODULE_ALIAS("ipt_CHAOS");
